"""
   Copyright 2011 Chris Read (chris.read@gmail.com)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import logging
import os
import types
import sys

try:
    from pysnmp.entity.rfc3413.oneliner import cmdgen
except ImportError:
    print "ERROR: Unable to load 'pysnmp' module."
    sys.exit(1)


class Ecks():
    """
    A simple way to get data out of a remote machine using SNMP without having to deal with a single MIB or OID
    
    Simple Usage:

    >>> import ecks
    >>> e = ecks.Ecks()
    >>> e.get_data('127.0.0.1', 'public', 'disk')
    [(2, 'Physical memory', 8589934592, 5169360896), (3, 'Swap space', 134213632, 45056), (4, '/', 290984034304, 243201781760)]
    >>> e.get_data('127.0.0.1', 'public', 'cpu')
    (21, 9, 68)
    >>> e.get_data('127.0.0.1', 'public', 'uptime')
    18879153


    """

    plugins = []

    def __init__(self, timeout = 1):
        self.timeout = timeout
        if logging._handlers == {}:
            logging.basicConfig()
        self.logger = logging.getLogger(__name__ + ".Ecks")
        self._load_plugins()


    def _load_plugins(self):
        """
        Walk through our plugins and wire their get_ methods 
        """
        plugin_dir = os.path.join(os.path.dirname(__file__), "plugins")
        for plugin in os.listdir(plugin_dir):
            if plugin.endswith(".py") and (plugin != "__init__.py"):
                self.logger.info("Loading file: %s" % os.path.join(plugin_dir, plugin))
                plugin_name = plugin[:-3]
                try:
                    exec("import plugins.%s" % plugin_name)
                    exec("self.get_%(func)s = types.MethodType(plugins.%(func)s.get_%(func)s, self)" % {"func": plugin_name})
                    self.logger.info("Registered plugin: %s" % plugin_name)
                    self.plugins += [plugin_name]
                except AttributeError as e:
                    self.logger.warn("Invalid plugin: %s" % plugin_name)


    def _extract(self, data, value_type, filter):
        """
        Return a filtered list of data that matches the given filter and is coerced to type value_type

        data
            A sequence in the form (oid, (data_type, index) value) as returned by get_snmp_data

        value_type
            A primitive python type to coerce the value into from the pysnmp type it is

        filter
            An integer defining what oid type to return
        """
        return [ value_type(value) for (oid, (data_type, index), value) in data if data_type == filter ]


    def _build_answer(self, *answers):
        return tuple([ a for a in answers ])

    def get_snmp_data(self, host, community, query_oid, query_oid_only = None):
        """
        Get data from server using a Bulk SNMP Get.

        host
            Hostname of the machine you want to pull the data from

        community
            SNMP community string to use

        query_oid
            If set to something that evaluates as True then it will only return
            results that match the query oid. By default it will return everything.
        """
        error_indication, error_status, error_index, var_binds_list = cmdgen.CommandGenerator().bulkCmd(
            cmdgen.CommunityData(host, community),
            cmdgen.UdpTransportTarget((host, 161), timeout = self.timeout),
            0, 25, query_oid)

        if error_indication:
            self.logger.error(error_indication)
        elif error_status:
            self.logger.error(
                '%s at %s\n' % (error_status.prettyPrint(), error_index and var_binds_list[int(error_index) - 1] or '?'))

        prefix = len(query_oid)
        data = [ (tuple(oid)[:prefix], tuple(oid)[prefix:], val) for [(oid, val)] in var_binds_list ]
        if query_oid_only:
            data = [ (oid, key, value) for (oid, key, value) in data if oid == query_oid ]

        return data


    def get_data(self, host, comm, plugin):
        """
        Utility method to interface with plugins by name

        host
            Hostname of the machine you want to pull the data from

        community
            SNMP community string to use

        plugin
            The plugin to call
        """
        return eval("self.get_%s(host, comm)" % plugin)



"""
   Ecks plugin to collect system uptime

   Copyright 2011 Chris Read (chris.read@gmail.com)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import re

def get_uptime(parent, host, community):
    """ This is a plugin to be loaded by Ecks

    return the number of TimeTicks (100ths of a second) since the system was booted

    Note: This plugin now corrects the bad uptime value returned in
          Windows builds < 7600 (Win 2008 R2)
          
          See the following for more info:
            http://fixunix.com/snmp/64367-re-interpreting-hrsystemuptime.html
            http://forums.cacti.net/about22137.html&highlight=
            http://web.archiveorange.com/archive/v/WnPpStN7UOuJhcUk5J0T

    """
    uptime = (1,3,6,1,2,1,25,1,1) # HOST-RESOURCE-MIB
    data = parent.get_snmp_data(host, community, uptime, 1)

    if data:
        # Got some data, let's check the OS now...   
        os = parent.get_os(host, community)
        # If it's Windows, check the build...
        if ("Windows" in os) and (int(re.search("\(Build (\d+) ", os).groups()[0]) < 7600):
            return int(data[0][2]) / 10
        else:
            return int(data[0][2])
    else:
        return None




"""
    A simple way to get data out of a remote machine using SNMP without having to deal with a single MIB or OID

    The goal of Ecks is simple - make it really easy to get get any data
    from an SNMP service.

    Ecks is made up of a core class that will collect data via SNMP,
    and a set of plugins that contain the OID and the code needed to
    transform the results from nested OID's to usable data.

    See help(ecks.Ecks) for more info


    Copyright 2011 Chris Read (chris.read@gmail.com)

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

__version__ = '0.5'
__author__ = 'Chris Read <chris.read@gmail.com>'

from ecks import Ecks



"""
   Ecks plugin to collect system memory usage information

   Copyright 2011 Chris Read (chris.read@gmail.com)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

def get_memory(parent, host, community):
    """ This is a plugin to be loaded by Ecks

    return a tuple containing (total_swap, avail_swap, total_real, avail_real). Values are in kiloBytes

    """
    memory = (1,3,6,1,4,1,2021,4) # UCD-SNMP-MIB
    data = parent.get_snmp_data(host, community, memory, 1)

    if data:
        return map(parent._build_answer,
            parent._extract(data, int, 3),
            parent._extract(data, int, 4),
            parent._extract(data, int, 5),
            parent._extract(data, int, 6),
        )[0]



"""
Installs ecks using setuputils

Run:

  python setup.py install

to install this package using this script. 

To install the latest public release use run:

  pip install ecks
"""

try:
  from setuptools import setup, find_packages
except:
  from distutils.core import setup, find_packages

import sys

import ecks

required_python_version = '2.6'

if sys.version < required_python_version:
  print "Ecks requires python %s or later, you have %s" % (required_python_version, sys.version)
  sys.exit(1)

setup(
  name = "Ecks",
  version = ecks.__version__,
  description = "Easy access to SNMP data",
  long_description = "A simple way to get data out of a remote machine using SNMP without having to deal with a single MIB or OID.",
  classifiers = [
    "Development Status :: 4 - Beta",                                                                                             
    "Intended Audience :: Developers",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Topic :: Internet",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: System :: Monitoring"
  ],
  author = "Chris Read",
  author_email = "chris.read@gmail.com",
  url = "http://github.com/cread/ecks",
  license = "Apache 2.0",
  packages = find_packages(),
  keywords = "snmp monitoring",
  install_requires = "pysnmp >= 4.1.8a"
)




"""
   Ecks plugin to collect TCP stack statistics

   Copyright 2011 Chris Read (chris.read@gmail.com)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

def ip_addr(data):
    return ".".join([ str(a) for a in data ])

def get_tcpstats(parent, host, community):
    """ This is a plugin to be loaded by Ecks

    return a list containing [
        tcpAttemptFails,    - http://www.oid-info.com/get/1.3.6.1.2.1.6.7
        tcpEstabResets,     - http://www.oid-info.com/get/1.3.6.1.2.1.6.8
        tcpCurrEstab        - http://www.oid-info.com/get/1.3.6.1.2.1.6.9
        tcpInSegs,          - http://www.oid-info.com/get/1.3.6.1.2.1.6.10
        tcpOutSegs,         - http://www.oid-info.com/get/1.3.6.1.2.1.6.11
        tcpRetransSegs,     - http://www.oid-info.com/get/1.3.6.1.2.1.6.12
        tcpInErrs,          - http://www.oid-info.com/get/1.3.6.1.2.1.6.14
        tcpOutRsts,         - http://www.oid-info.com/get/1.3.6.1.2.1.6.15
    ]

    """
    ids = [7, 8, 9, 10, 11, 12, 14, 15]

    stats = []
    for id in ids:
        try:
            stats = stats + [long(parent.get_snmp_data(host, community, (1,3,6,1,2,1,6,id), 1)[0][2])]
        except:
            return

    return stats
